; *****************************************************************************         
;               MidiSwap.asm
;       A simple MIDI splitter/patch bay using a PIC16F876
;       Davide Bucci, 2004-2007
;       Version 2.0
; *****************************************************************************
;
;License:
; --------
;
;    Copyright (C) 2004-2007  Davide Bucci  davbucciPleasenospamherE@tiscali.it
;
;    This program is free software; you can redistribute it and/or modify
;    it under the terms of the GNU General Public License as published by
;    the Free Software Foundation; either version 2 of the License, or
;    (at your option) any later version.
;
;    This program is distributed in the hope that it will be useful,
;    but WITHOUT ANY WARRANTY; without even the implied warranty of
;    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;    GNU General Public License for more details.
;
;    You should have received a copy of the GNU General Public License
;    along with this program; if not, write to the Free Software
;    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

        PROCESSOR 16F876
        ERRORLEVEL -302         ; disable the annoying warning 
                                ; "Register in operand not in bank 0. 
                                ; Ensure bank bits are correct."


        #include "p16f876.inc"
        __CONFIG 0x3F7A         ; Disable watchdog, HS oscillator, 
                                ; disable LT programming

; Control lines of the display device
E       equ     01
RW      equ     03
RS      equ     02
; Control port used for LCD display
CNTRLCD equ     PORTB
; Data port used for LCD display
DATALCD equ     PORTB
TRISLCD equ     TRISB

; These registers must be in the same page of the CNTRLCD register              
TMP     equ     20      ; Dummy for LCD nibble mode
CNT     equ     21      ; Counter
SDR     equ     22      ; Short Delay Register
LDR     equ     23      ; Long Delay Register

; These registers controls the buttons
BUTTONS equ     PORTA   
SET_B   equ     0
CHNG_B  equ     1
OLD_BTN equ     28
ACT_BTN equ     29
TMP_1   equ     2A
HND     equ     2B
DEC     equ     2C
UNT     equ     2D


; These registers are accessible in all pages
STORES  equ     70      ; Store status register
STOREW  equ     71      ; Store the w register
STOREP  equ     72      ; Store the PCLATH register

; *****************************************************************************         
;  	CHANGE HERE THE NUMBER OF OUTPUTS ACCORDINGLY TO YOUR HARDWARE
; *****************************************************************************  

NOPORTS equ    6        ; Number of outputs in the bay (min: 2, max 6)

TRNSM   equ     26      ; Transmission status (See TRN section)
RECBYTE equ     27      ; Received byte (used in interrupt)
TRN_RXF equ     0       ; Error: reception buffer full
TRN_FRM equ     1       ; Error: reception without a stop bit (frame error)

ACT_CHN equ     2E      ; Actual channel being modified (0-15)
ACT_POR equ      2F      ; Actual port being modified (0-NOPORTS-1)
MIDI_T  equ     30      ; MIDI channel table (16 bytes starting from here)
TMP_2   equ     41
CNT_1   equ     42
CNT_2   equ     43

; *****************************************************************************         
;               MACROS
; *****************************************************************************         
                ; Write a message on the LCD display
WRITELN         MACRO   msg
                local   loop_ch
                local   end_mes
                BANKSEL CNT
                clrf    CNT
loop_ch         
                
                movlw   HIGH msg
                movwf   PCLATH
                movlw   LOW msg
                addwf   CNT,w
                movlw   HIGH msg
                btfsc   STATUS,C
                addlw   1
                movwf   PCLATH
                movf    CNT,w
                call    msg
                xorlw   0x00
                btfsc   STATUS,Z        ; Test if the character is a zero
                goto    end_mes
                bsf     CNTRLCD,RS      ; RS to 1: we are sending a character
                bcf     CNTRLCD,RW      ; RW to 0: we are writing to display
                call    sendbyte
                incf    CNT, f
                goto    loop_ch
end_mes 
                ENDM

; *****************************************************************************         
;               Main Program
; *****************************************************************************         
                org     0000
        
                goto    prg
                nop     
                nop
                nop
                org     0004
                goto    interrupt


prg             
                BANKSEL ADCON1
                movlw   0x06
                movwf   ADCON1          ; Disable the ADC on port A

                BANKSEL TRISB
                clrf    CNTRLCD         ; Display ports as outputs
                clrf    DATALCD 
                bcf     TRISC, 0        ; Out port A selection 
                bcf     TRISC, 1        ; Out port B selection
                bcf     TRISC, 2        ; Out port C selection 
                bcf     TRISC, 3        ; Out port D selection
                bcf     TRISC, 4        ; Out port E selection 
                bcf     TRISC, 5        ; Out port F selection
                bsf     BUTTONS, SET_B  ; Button lines as inputs
                bsf     BUTTONS, CHNG_B
                BANKSEL PORTB
        
        
                call    longdelay       ; Wait...
                call    longdelay
                call    longdelay

                movlw   MIDI_T          ; Read the MIDI table from Data EEPROM
                movwf   FSR
                clrf    TMP             ; Use TMP as counter (no problem 
                                      ; outside the LCD routines)
next            movfw   TMP             ; TMP is the address used for EEPROM 
                                  ; reading
                BANKSEL EEADR
                movwf   EEADR
                BANKSEL EECON1
                bcf     EECON1, EEPGD   ; Select the data EEPROM
                bsf     EECON1, RD      ; Select reading from the data EEPROM
                BANKSEL EEDATA
                movfw   EEDATA
                movwf   INDF
                incf    FSR,f
                BANKSEL TMP
                incf    TMP,f
                btfss   TMP, 4          ; Test the fourth bit: 16 locations
                goto    next
                
                call    functionset
                call    displayon


                BANKSEL ACT_CHN
                clrf    ACT_CHN
                                        ; UART settings for MIDI transmission
                BANKSEL TRNSM
                clrf    TRNSM           ; Clear the transmission error register

                movlw   7               ; Set transmission speed (31250 bauds 
                                        ; for a 4MHz crystal)
                BANKSEL SPBRG
                movwf   SPBRG
                BANKSEL TXSTA
                

                
                
                bcf     TXSTA, TX9      ; 8 bits transmission
                bcf     TXSTA, SYNC     ; Asynchronous mode
                bsf     TXSTA, BRGH     ; High speed mode
                bsf     TXSTA, TXEN     ; Enable transmission

                BANKSEL PIE1
                bsf     PIE1, RCIE      ; Use interrupt driven reception
                BANKSEL RCSTA
                bsf     RCSTA, SPEN     ; Serial port enabled
                bcf     RCSTA, RX9      ; 8 bits reception
                bsf     RCSTA, CREN     ; Continous receive on
                bcf     RCSTA, ADDEN
                bcf     RCSTA, FERR


                bsf     INTCON, PEIE    ; Set USART interrupt enable

                bsf     INTCON, GIE     ; Set Global Interrupt enable
                
                bsf     PORTC, 0        ; Enable Midi Out A

                call    displayclear            
                call    displaychome

                WRITELN text_welcome    ; Greetings and program version.
                movlw   0x40
                call    displayaddrset  ; Move to the second line
                WRITELN text_davide     ; Copyright

                call    longdelay       ; Wait...
                call    longdelay
                call    longdelay

                call    longdelay       ; Wait...
                call    longdelay
                call    longdelay

                goto    state_split     ; The selection mode is the fist state 
                                        ; of the state machine          

; *****************************************************************************
        

signal_rxf      call    displayclear    ; A buffer overrun occourred, write an 
                                        ; error message
                call    displaychome
                WRITELN text_err_bor
                bcf     TRNSM, TRN_RXF
                call    longdelay
                call    longdelay
                call    longdelay
                return

signal_frm      call    displayclear    ; A frame error occourred, write an 
                                   ; error message
                call    displaychome
                WRITELN text_err_frm
                bcf     TRNSM, TRN_FRM
                call    longdelay
                call    longdelay
                call    longdelay
                return
                
wait_buttons                            ; Wait until a button is pressed; 
                                        ; return the button code in w and in 
                                        ; OLD_BTN
                movlw   0x20            ; Set the delay duration
                movwf   LDR             
                call    longdelay
button_loop                             ; Main loop
                BANKSEL BUTTONS
                movfw   ACT_BTN         ; Save the old state of the buttons
                movwf   OLD_BTN 
        
                BANKSEL TRNSM   
                btfsc   TRNSM, TRN_RXF  ; Test if a buffer overrun condition 
                                        ; occourred
                call    signal_rxf
                BANKSEL TRNSM
                btfsc   TRNSM, TRN_FRM  ; Test if a frame error occourred
                call    signal_frm
        
                BANKSEL BUTTONS

                movfw   BUTTONS         ; Save the new state of the buttons
                movwf   ACT_BTN
                xorwf   OLD_BTN,f       ; Verify if there was a change in the 
                                        ; button status
                btfsc   STATUS, Z
                goto    button_loop
                btfsc   ACT_BTN, SET_B  ; Test the set button
                retlw   SET_B
                btfsc   ACT_BTN, CHNG_B ; Test the change button
                retlw   CHNG_B
                goto    wait_buttons


eeprom_write                            ; Toggle the port value of the current 
                                        ; MIDI channel and port
                BANKSEL MIDI_T
                movlw   MIDI_T          ; Load the address of the MIDI channel 
                                        ; table
                addwf   ACT_CHN,w       ; Look at the active channel
                movwf   FSR

                movfw   ACT_POR         ; Obtain a 1 in the bit position
                movwf   CNT_2           ; correspondng to the actual port
                incf    CNT_2,f         ; Compensate the first cycle
                clrf    TMP_2
                bsf     STATUS,C        ; Cycle through carry
looprotate      rlf     TMP_2,f         
                decfsz  CNT_2,f
                goto    looprotate
             
               
                movfw   TMP_2           ; Here w has a 1 in the right position
                xorwf   INDF,f          ; Modify the corresponding port status

                BANKSEL EECON1
                btfsc   EECON1, WR      ; Wait for
                goto    $-1             ; write to finish
                BANKSEL ACT_CHN
                movfw   ACT_CHN         ; Load the address
                BANKSEL EEADR
                movwf   EEADR           
                movfw   INDF
                movwf   EEDATA          ; Put in EEDATA the value in the table
                BANKSEL EECON1
                bcf     EECON1, EEPGD   ; Point to Data memory
                bsf     EECON1, WREN    ; Enable writes
                                        
                bcf     INTCON, GIE     ; Disable interrupt
                movlw   0x55            ; Write 55h to
                movwf   EECON2          ; EECON2
                movlw   0xAA            ; Write AAh to
                movwf   EECON2          ; EECON2
                bsf     EECON1, WR      ; Start write operation
                bsf     INTCON, GIE     ; Enable interrupts
                bcf     EECON1, WREN    ; Disable writes
                return

; *****************************************************************************         
;               User interface state machine
; *****************************************************************************         

state_split                             ; Splitting state: does nothing but 
                                        ; waiting user interaction
                clrf    ACT_POR 
                call    displayclear
                call    displaychome
                WRITELN text_split      ; Print "Splitting..."
                movlw   0x00
                call    displayaddrset  ; Move to the first line
                call    wait_buttons
                btfsc   ACT_BTN, SET_B
                goto    state_sel_menu  ; Jump to the state "selection menu"
                goto    state_split

state_sel_menu                          ; Selection mode menu
                clrf    ACT_POR 
                call    print_sel_menu
                movlw   0x00
                call    displayaddrset  ; Move to the first line
                call    wait_buttons
                btfsc   ACT_BTN, SET_B
                goto    state_split     ; Jump to the state "splitting"
                btfsc   ACT_BTN, CHNG_B
                goto    state_sel_chan  ; Jump to the state "selection channel"
                goto    state_sel_menu

state_sel_chan                          ; Selection channel menu
                call    print_sel_menu
                movlw   0x44
                call    displayaddrset  ; Move to the second line
                call    wait_buttons
                btfsc   ACT_BTN, SET_B  ; If the set button is pressed
                incf    ACT_CHN, f      ; Increment channel number
                btfsc   ACT_CHN, 4      ; If ACT_CHN==16, clear it
                clrf    ACT_CHN
                btfsc   ACT_BTN, CHNG_B
                goto    state_sel_port  ; Jump to the state "change port"
                goto    state_sel_chan

state_sel_port                          ; Selection port menu
                call    print_sel_menu
                movlw   0x4A
                addwf   ACT_POR,w       ; Position of the highlighted port
                call    displayaddrset  ; 
                call    wait_buttons
        
                btfsc   ACT_BTN, SET_B  ; If the set button is pressed
                call    eeprom_write    ; Save the configuration in EEPROM

next_button     btfss   ACT_BTN, CHNG_B
                goto    state_sel_port
                incf    ACT_POR,f
                movfw   ACT_POR
                xorlw   NOPORTS
                btfsc   STATUS,Z        ; Test if we reached the last port
                goto    state_sel_menu  ; Jump to the state "change port"
                goto    state_sel_port


print_sel_menu                          ; Print the selection mode menu
                call    displayclear
                call    displaychome
                WRITELN text_selection  ; Print "Selection Mode"
                movlw   0x40
                call    displayaddrset  ; Move to the second line
                WRITELN text_ch         ; Print "Ch. "
                movfw   ACT_CHN
                addlw   1       
                call    write_number    ; Write the number
                movlw   0x47
                call    displayaddrset
                WRITELN text_out        ; Print "Out "          
                movlw   MIDI_T          ; Load the address of the MIDI channel 
                                        ; table
                addwf   ACT_CHN,w       ; Look at the active channel
                movwf   FSR
                movfw   INDF            ; Load the state of the active channel
                movwf   TMP_2           ; save the state in a temp file
                movlw   NOPORTS
                movwf   CNT_1           ; counter: for each port
                clrf    CNT_2
loopports       movfw   CNT_2
                btfss   TMP_2,0         
                goto skipport
                addlw   'A'             ; The port is active
                bsf     CNTRLCD,RS      ; RS = 1: we are sending a character
                bcf     CNTRLCD,RW      ; RW = 0: we are writing to the display
                call    sendbyte        ; Send the channel number
                goto    loopportsout
skipport
                movlw   0xA5            ; The port is not active (dot)
                bsf     CNTRLCD,RS      ; RS = 1: we are sending a character
                bcf     CNTRLCD,RW      ; RW = 0: we are writing to the display
                call    sendbyte        ; Send the channel number

loopportsout    incf    CNT_2,f
                rrf     TMP_2,1         
                decfsz  CNT_1,f
                goto    loopports
                return

; *****************************************************************************         
;               Interrupt handling routine
; *****************************************************************************         

interrupt       
                movwf   STOREW          ; Context save
                movfw   STATUS
                movwf   STORES
                movfw   PCLATH
                movwf   STOREP
                

charloop        
                BANKSEL PIR1            ; Loop until there is a char
                                   ; in the recption buffer
                btfss   PIR1, RCIF      ; Character received?
                goto    exitinterrupt   ; All characters processed
                
                BANKSEL RCREG   
                movfw   RCREG           ; Read character

                ; Recognise the start of a MIDI message is a pretty easy task
                ; if we note that it's the only byte in the transmission having
                ; the MSB set. The following bytes will be treated as part of 
                ; the  same message.
                ; In all MIDI message, the higher four bits in the first byte 
                ; specifies the message type and the four lower the channel.
                ; In our case, we simply read the allocation table at the 
                ; position given by the channel, to activate the output
                ; ports.
                
                BANKSEL RECBYTE         ; Save the received byte (same bank
                                    ; of PORTC)
                movwf   RECBYTE
                btfss   RECBYTE, 7      ; Test the most significative byte
                goto    skip            ; Is not a MIDI message header
                
                ; This code is executed only for the MIDI message header.
                ; In this place we should recognise the channel and decide
                ; port should be activated when the message will be sent.
                ; TODO: recognize non-channel messages (such as SYSEX)

                andlw   0x0F            ; Separate the channel number

                addlw   MIDI_T          ; Look in the table for the channel
                movwf   FSR
                movfw   INDF            ; Load the state of the active channel
                movwf   TMP

portA           bcf     PORTC, 0
                btfsc   TMP,0           ; test the 0 bit: port A
                bsf     PORTC, 0
portB           bcf     PORTC, 1
                btfsc   TMP,1           ; test the 0 bit: port B
                bsf     PORTC, 1
portC           bcf     PORTC, 2
                btfsc   TMP,2           ; test the 0 bit: port C
                bsf     PORTC, 2
portD           bcf     PORTC, 3
                btfsc   TMP,3           ; test the 0 bit: port D
                bsf     PORTC, 3
portE           bcf     PORTC, 4
                btfsc   TMP,4           ; test the 0 bit: port E
                bsf     PORTC, 4
portF           bcf     PORTC, 5
                btfsc   TMP,5           ; test the 0 bit: port F
                bsf     PORTC, 5

                ; Send the byte received        
skip            movfw   RECBYTE
                movwf   TXREG
        
                BANKSEL PIR1                    
waitendtr       btfss   PIR1, TXIF
                goto    waitendtr       ; Wait the end of transmission of the 
                                        ; previous byte
                
                btfsc   TXSTA, OERR     ; Test if a buffer overrun occourred
                call    error_buffer
                btfsc   TXSTA, FERR     ; Test if a frame error occourred
                call    error_frame     
                goto    charloop        ; See if there is another byte pending          
                                        ; The RCIF bit is cleared by hardware 
                                        ; by reading the RCREG register
exitinterrupt   
                movfw   STOREP          ; Context restore
                movwf   PCLATH
                movfw   STORES                  
                movwf   STATUS
                movfw   STOREW
                retfie                  ; Return from interrupt

error_buffer    
                BANKSEL TRNSM           ; Error condition: buffer overrun
                bsf     TRNSM, TRN_RXF  
                BANKSEL RCSTA
                bcf     RCSTA, CREN     ; Cont. receive off: reset the USART
                bsf     RCSTA, CREN     ; Cont. receive on: restart 
                return

error_frame     
                BANKSEL TRNSM           ; Error condition: frame err
                bsf     TRNSM, TRN_FRM          
                BANKSEL RCSTA
                bcf     RCSTA, CREN     ; Cont. receive off: reset the USART
                bsf     RCSTA, CREN     ; Cont. receive on: restart 
                return


; *****************************************************************************         
;               LCD Subroutines 
; *****************************************************************************         
write_number                            ; Print the number in w 
                BANKSEL TMP_1
                movwf   TMP_1           ; Save the w register in memory
                movlw   -1
                movwf   HND
                movfw   TMP_1
sub_hnd         addlw   -0x64           ; Substract 100(10)
                incf    HND, f          ; Hundreds
                btfsc   STATUS, C       ; Verify if the result is negative
                goto    sub_hnd         
                
                addlw   0x64            ; Add 100(base10) (to compensate the 
                                        ; last subtraction)     
                movwf   TMP_1           ; Save the w register in memory
                movlw   -1
                movwf   DEC
                movfw   TMP_1
sub_dec         addlw   -0xA            ; Substract 10(base10)
                incf    DEC, f          ; 
                btfsc   STATUS, C       ; Verify if the result is negative
                goto    sub_dec         

                addlw   0x0A            ; Add 10(base10) (to compensate the 
                                        ; last subtraction)             
                movwf   UNT             ; Save the units

                
                bcf     TMP_1,0         ; Flag: avoid unnecessary zeros
                movfw   HND             ; Send the hundreds
                addlw   0x00            ; Used only to test if w is zero (does 
                                        ; not modify w)
                btfsc   TMP_1, 0        ; Force to write all numbers
                goto    force_HND
                btfsc   STATUS, Z       ; Skip if necessary
                goto    skip_HND
force_HND       addlw   '0'             ; Transform it in ASCII
                bsf     CNTRLCD,RS      ; RS to 1: we are sending a character
                bcf     CNTRLCD,RW      ; RW to 0: we are writing to display
                call    sendbyte        ; Send the channel number               
                bsf     TMP_1, 0        ; Force to write successive zeros
skip_HND        movfw   DEC             ; Send the tenths
                addlw   0x00            ; Used only to test if w is zero 
                                        ; (does not modify w)
                btfsc   TMP_1, 0        ; Force to write all numbers
                goto    force_DEC
                btfsc   STATUS, Z       ; Skip if necessary
                goto    skip_DEC
force_DEC       addlw   '0'             ; Transform it in ASCII
                bsf     CNTRLCD,RS      ; RS to 1: we are sending a character
                bcf     CNTRLCD,RW      ; RW to 0: we are writing to display
                call    sendbyte        ; Send the channel number
                bsf     TMP_1, 0
skip_DEC        movfw   UNT             ; Send the units (always to be written)
                addlw   '0'             ; Transform it in ASCII
                bsf     CNTRLCD,RS      ; RS to 1: we are sending a character
                bcf     CNTRLCD,RW      ; RW to 0: we are writing to display
                call    sendbyte        ; Send the channel number
                return


functionset     
                BANKSEL CNTRLCD
                bcf     CNTRLCD,RS
                bcf     CNTRLCD,RW
                movlw   0x20            ; Nibble mode of the LCD
                movwf   DATALCD
                call    pulse_e
                call    busywait
functionset2    bcf     CNTRLCD,RS
                bcf     CNTRLCD,RW
                movlw   0x28
                call    sendbyte                
                call    busywait
                retlw   0

                ; Send the command in w to the display LCD configured in the 
                ; nibble mode
sendbyte        
                BANKSEL CNTRLCD
                movwf   TMP             ; Store command temporarily in TMP
                call    portnibble             
                call    pulse_e         ; Send the first half of the command
                swapf   TMP,w           ; swap nibbles of TMP, result in w
                call    portnibble              
                call    pulse_e         ; Send the second half of the command
                call    busywait
                retlw   0

                ; Activate the display, cursor underline blinking
displayon      
                BANKSEL CNTRLCD
                bcf     CNTRLCD,RS
                bcf     CNTRLCD,RW
                movlw   0x0D
                call    sendbyte
                retlw   0
                
                ; Clear the display
displayclear    
                BANKSEL CNTRLCD         
                bcf     CNTRLCD,RS
                bcf     CNTRLCD,RW
                movlw   0x01
                call    sendbyte
                retlw   0
                
                ; Put the cursor at home
displaychome    
                BANKSEL CNTRLCD
                bcf     CNTRLCD,RS
                bcf     CNTRLCD,RW
                movlw   0x02
                call    sendbyte
                retlw   0

                
                ; Set cursor address (passed in w)
displayaddrset  
                BANKSEL CNTRLCD
                bcf     CNTRLCD,RS
                bcf     CNTRLCD,RW
                iorlw   0x80
                call    sendbyte
                retlw   0

                ; w choose method: bit 0=0 decrement =1 increment; 
                ; bit 1=0 display shift off =1 display shift on
displaycurdir   
                BANKSEL CNTRLCD
                bcf     CNTRLCD,RS
                bcf     CNTRLCD,RW
                iorlw   0x04
                call    sendbyte
                retlw   0

; *****************************************************************************         
;               Delay routines for LCD
; *****************************************************************************         
longdelay
                BANKSEL LDR
                call    shortdelay
                decfsz  LDR,f
                goto    longdelay
                return
                
shortdelay      
                BANKSEL SDR
                decfsz  SDR,f
                goto    shortdelay
                return

busywait        
                BANKSEL TRISLCD
                movlw   0xF0            ; Set all the port (4 bits) as inputs
                movwf   TRISLCD
                BANKSEL DATALCD
                bcf     DATALCD, RS     ; RS to 0
                bsf     DATALCD, RW     ; RW to 1: Read
                nop
                bsf     DATALCD, E      ; Raise E line
                nop     
                rlf     DATALCD, w      ; Rotate the busy flag in the carry
                bcf     DATALCD, E
                nop
                nop     
                bsf     DATALCD, E      ; Raise E line
                nop
                nop
                nop     
                bcf     DATALCD, E
                btfsc   STATUS, C       ; Test the carry
                goto    busywait        ; If busy, continue waiting
                BANKSEL TRISLCD
                movlw   0x00            ; Set all the port (4 bits) as outputs
                movwf   TRISLCD
                BANKSEL DATALCD
                return 
                

pulse_e         
                BANKSEL CNTRLCD
                bsf     CNTRLCD,E
                nop
                nop
                nop
                bcf     CNTRLCD,E
                return

portnibble      andlw   0xF0            ; clear the lower 4 bits of w
                BANKSEL DATALCD
                iorwf   DATALCD, f      ; or this with port B
                iorlw   0x0F            ; set lower 4 bits of w
                andwf   DATALCD, f      ; and this with port B
                return
                
; *****************************************************************************         
;               Text tables
; *****************************************************************************         

text_welcome    addwf   PCL,f
                DT      "Welcome   V. 2.0",0
text_davide        addwf   PCL,f
          DT      "D. Bucci 2004-07"
text_err_frm    addwf   PCL,f
                DT      "MIDI Frame Err.",0
text_err_bor    addwf   PCL,f
                DT      "MIDI Buffer Err.",0            
text_split      addwf   PCL, f
                DT      "Splitting...",0
text_selection  addwf   PCL, f
                DT      "Selection Mode",0
text_ch         addwf   PCL, f
                DT      "Ch. ",0
text_out        addwf   PCL, f
                DT      "P. ",0
                end
